from __future__ import annotations

import json
import sys
import uuid
from collections import Counter
from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Set, Union

from pydantic import (
    BaseModel,
    DirectoryPath,
    Extra,
    FilePath,
    parse_raw_as,
    root_validator,
    validator,
)
from storage_device_managers import ValidCompressions

FoldersT = Dict[DirectoryPath, str]


def path_aware_btrfs_json_decoding(v, *, default) -> str:
    v["Folders"] = {str(key): val for key, val in v["Folders"].items()}
    return json.dumps(v, default=default)


def path_aware_restic_json_decoding(v, *, default) -> str:
    v["FilesAndFolders"] = {str(cur) for cur in v["FilesAndFolders"]}
    return json.dumps(v, default=default)


class BaseConfig(BaseModel):
    BackupRepositoryFolder: str
    DevicePassCmd: str
    ExcludePatternsFile: Optional[FilePath] = None
    UUID: uuid.UUID
    Compression: Optional[ValidCompressions] = None


class BtrFSRsyncConfig(BaseConfig):
    Files: Set[FilePath]
    FilesDest: str
    Folders: FoldersT
    SubvolTimestampFmt: ClassVar[str] = "%F_%H:%M:%S"

    class Config:
        extra = Extra.forbid
        frozen = True
        json_dumps = path_aware_btrfs_json_decoding

    @validator("Files")
    def source_file_names_must_be_unique(cls, files):
        file_names = Counter(f.name for f in files)
        cls.raise_with_message_upon_duplicate(file_names, ("Dateinamen", "Dateinamen"))
        return files

    @validator("Folders")
    def folder_destinations_must_be_unique(cls, folders: FoldersT) -> FoldersT:
        destinations = Counter(folders.values())
        cls.raise_with_message_upon_duplicate(
            destinations, ("Zielverzeichnissen", "Ziele")
        )
        return folders

    @validator("ExcludePatternsFile", pre=True)
    def expand_tilde_in_exclude_patterns_file_name(cls, maybe_exclude_patterns):
        if maybe_exclude_patterns is None:
            return None
        return Path(maybe_exclude_patterns).expanduser()

    @validator("Files", pre=True)
    def expand_tilde_in_file_sources(cls, files):
        new = [Path(cur).expanduser() for cur in files]
        return new

    @validator("Folders", pre=True)
    def expand_tilde_in_folder_sources(cls, folders):
        new = {Path(src).expanduser(): dest for src, dest in folders.items()}
        return new

    @root_validator(skip_on_failure=True)
    def files_dest_is_no_folder_dest(cls, values):
        files_dest = values["FilesDest"]
        destinations = values["Folders"].values()
        if files_dest in destinations:
            raise ValueError(
                f"Zielverzeichnis {files_dest} ist gleichzeitig Ziel für Ordner und Einzeldateien."
            )
        return values

    @staticmethod
    def raise_with_message_upon_duplicate(
        counts: Union[Counter[Path], Counter[str]], token: tuple[str, str]
    ) -> None:
        if all(val == 1 for val in counts.values()):
            return
        errmsg_begin = (
            f"Duplikate in {token[0]} entdeckt. Folgende {token[1]} kommen doppelt vor:"
        )
        errmsg_body = " ".join(
            str(elem) for (elem, count) in counts.items() if count > 1
        )
        raise ValueError(f"{errmsg_begin} {errmsg_body}")

    def device(self) -> Path:
        return Path(f"/dev/disk/by-uuid/{self.UUID}")

    def map_name(self) -> str:
        return str(self.UUID)


class ResticConfig(BaseConfig):
    FilesAndFolders: Set[Union[FilePath, DirectoryPath]]
    RepositoryPassCmd: str

    class Config:
        extra = Extra.forbid
        frozen = True
        json_dumps = path_aware_restic_json_decoding

    @validator("ExcludePatternsFile", pre=True)
    def expand_tilde_in_exclude_patterns_file_name(cls, maybe_exclude_patterns):
        if maybe_exclude_patterns is None:
            return None
        return Path(maybe_exclude_patterns).expanduser()

    @validator("FilesAndFolders", pre=True)
    def expand_tilde_in_sources(cls, files_and_folders):
        new = {Path(src).expanduser() for src in files_and_folders}
        return new

    def device(self) -> Path:
        return Path(f"/dev/disk/by-uuid/{self.UUID}")

    def map_name(self) -> str:
        return str(self.UUID)


Configuration = Union[BtrFSRsyncConfig, ResticConfig]


def parse_configuration(content: str) -> list[Configuration]:
    config_lst = parse_raw_as(List[Configuration], content)
    if len(config_lst) == 0:
        sys.exit("Leere Konfigurationsdateien sind nicht erlaubt.\n")
    return config_lst



#!/usr/bin/env python3
import enum
import json
import os
import sys
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Callable, Optional

import storage_device_managers as sdm
import typer
from loguru import logger

from . import __version__
from . import backup_backends as bb
from . import config_parser as cp
from .device_managers import (
    prepare_device_for_butterbackend,
    prepare_device_for_resticbackend,
)

app = typer.Typer()
DEFAULT_CONFIG_DIR = Path("~/.config/").expanduser()
DEFAULT_CONFIG_NAME = "butter-backup.cfg"


class ValidBackends(enum.Enum):
    restic = "restic"
    btrfs_rsync = "btrfs-rsync"


def get_default_config_path() -> str:
    config_dir = Path(os.getenv("XDG_CONFIG_HOME", DEFAULT_CONFIG_DIR))
    config_file = config_dir / DEFAULT_CONFIG_NAME
    return str(config_file)


def setup_logging(verbosity: int) -> None:
    # If no `-v/--verbose` is given (verbosity == 0 in this case), errors and
    # warnings shall appear. The first flag shall let successes appear, so that
    # the user can trace the progress of the program.
    logger.remove()
    available_levels = [
        # "CRITICAL",
        # "ERROR",
        "WARNING",
        "SUCCESS",
        "INFO",
        "DEBUG",
        "TRACE",
    ]
    level = min(verbosity, len(available_levels) - 1)
    logger.add(sys.stderr, level=available_levels[level])


CONFIG_OPTION = typer.Option(get_default_config_path(), exists=True, dir_okay=False)
VERBOSITY_OPTION = typer.Option(0, "--verbose", "-v", count=True)


@app.command()
def open(  # noqa: A001
    config: Path = CONFIG_OPTION, verbose: int = VERBOSITY_OPTION
) -> None:
    """
    Öffne alle in der Konfiguration gelisteten Speichermedien

    Das Kommando `open` öffnet alle Speichermedien, deren UUID in der
    Konfiguration erwähnt wird.  Für jedes geöffnete Speichermedium wird die
    UUID und der Mount-Zielordner angegeben.

    Dies Kommando ist besonders nützlich um Sicherheitskopien
    wiederherzustellen. Dafür wird das Speichermedium, auf dem sich die
    Sicherheitskopien befinden, mittels `butter-backup open` geöffnet. Dann
    kann mit den Daten interagiert werden, z.B. durch Öffnen im Dateibrowser
    oder durch Verwendung von `restic`. Nach erfolgreicher Wiederherstellung
    kann das Speichermedium mit `butter-backup close` wieder entfernt werden.
    """
    setup_logging(verbose)
    configurations = cp.parse_configuration(config.read_text())
    for cfg in configurations:
        if cfg.device().exists():
            mount_dir = Path(mkdtemp())
            decrypted = sdm.open_encrypted_device(cfg.device(), cfg.DevicePassCmd)
            sdm.mount_btrfs_device(
                decrypted, mount_dir=mount_dir, compression=cfg.Compression
            )
            typer.echo(f"Speichermedium {cfg.UUID} wurde in {mount_dir} geöffnet.")


@app.command()
def close(config: Path = CONFIG_OPTION, verbose: int = VERBOSITY_OPTION) -> None:
    """
    Schließe alle geöffneten Speichermedien

    Das Kommando `close` schließt alle gemounteten Speichermedien, deren UUIDs
    in der Konfiguration erwähnt werden. Es ist das Gegenstück des Kommandos
    `open`. Weitere Erklärungen finden sich dort.
    """
    setup_logging(verbose)
    configurations = cp.parse_configuration(config.read_text())
    mounted_devices = sdm.get_mounted_devices()
    for cfg in configurations:
        mapped_device = f"/dev/mapper/{cfg.UUID}"
        if cfg.device().exists() and mapped_device in mounted_devices:
            mount_dirs = mounted_devices[mapped_device]
            if len(mount_dirs) != 1:
                # TODO introduce custom exception
                raise ValueError(
                    "Got several possible mount points. Expected exactly 1!"
                )
            mount_dir = mount_dirs.pop()
            sdm.unmount_device(mount_dir)
            sdm.close_decrypted_device(Path(mapped_device))
            mount_dir.rmdir()


@app.command()
def backup(config: Path = CONFIG_OPTION, verbose: int = VERBOSITY_OPTION) -> None:
    """
    Führe Sicherheitskopien durch

    Für jedes angeschlossene Speichermedium wird eine Sicherheitskopie gemäß
    der Konfiguration durchgeführt. Das Speichermedium wird hierfür
    entschlüsselt und gemountet. Nachdem die Sicherungskopie durchgeführt
    wurde, wird es wieder vollständig geschlossen.

    Die Sicherheitskopien werden sequentiell durchgeführt. Dadurch wird
    sichergestellt, dass auch auf HDDs brauchbare Lesegeschwindigkeiten erzielt
    werden können.

    Direkt nach Beendigung der Durchführung der Sicherheitskopien, d.h. nachdem
    `butter-backup backup` zurückgekehrt ist oder nachdem die Durchführung der
    Sicherheitskopien des nächsten Speichermediums begonnen wurde, kann das
    entsprechende Speichermedium physisch entfernt werden. Eine Wartezeit oder
    weitere manuelle Schritte sind nicht nötig.
    """
    setup_logging(verbose)
    configurations = cp.parse_configuration(config.read_text())
    for cfg in configurations:
        if not cfg.device().exists():
            logger.info(
                f"Speichermedium {cfg.UUID} existiert nicht. Es wird kein Backup angelegt."
            )
            continue
        backend = bb.BackupBackend.from_config(cfg)
        with sdm.decrypted_device(cfg.device(), cfg.DevicePassCmd) as decrypted:
            with sdm.mounted_device(decrypted, cfg.Compression) as mount_dir:
                backend.do_backup(mount_dir)


@app.command()
def format_device(
    backend: ValidBackends = typer.Argument(...),  # noqa: B008
    device: Path = typer.Argument(  # noqa: B008
        ..., exists=True, dir_okay=False, readable=False
    ),
    config_to: Optional[Path] = typer.Option(  # noqa: B008
        None,
        help="Datei, in welche die generierte Konfiguration geschrieben werden"
        " soll. Die angegebene Datei darf nicht existieren. Wenn nicht"
        " angegeben, wird die Konfiguration auf STDOUT ausgegeben.",
    ),
    verbose: int = VERBOSITY_OPTION,
) -> None:
    """
    Richtet Speichermedium für butter-backup ein

    Das angegebene Speichermedium wird vollständig zur Erstellung von
    Sicherheitskopien mit `butter-backup` vorbereitet. Es wird eine
    Konfiguration ausgegeben, die nur noch um die zu sichernden Ordner bzw.
    Dateien ergänzt werden muss.

    Die in der ausgegebenen Konfiguration enthaltenen Passwörter werden mittels
    kryptographisch sicheren Methoden erstellt.

    Es wird dringend angeraten, die Passwörter nicht in der Konfiguration zu
    belassen, sondern in einen Passwortmanager zu tun. Der Autor verwendet
    `butter-backup` zusammen mit dem Passwortmanager `pass`.
    """
    setup_logging(verbose)
    config_writer: Callable[[str], Any]
    if config_to is None:
        config_writer = typer.echo
    else:
        if config_to.exists():
            raise ValueError(
                "Zieldatei für ButterBackup-Konfiguration existiert schon!"
            )
        config_writer = config_to.write_text
    formatter = (
        prepare_device_for_butterbackend
        if backend == ValidBackends.btrfs_rsync
        else prepare_device_for_resticbackend
    )
    config = formatter(device)
    json_serialisable = json.loads(config.json(exclude_none=True))
    config_writer(json.dumps([json_serialisable], indent=4, sort_keys=True))


@app.command()
def version() -> None:
    """Gibt butter-backups aktuelle Version an"""
    typer.echo(__version__)


def cli() -> None:
    app()


if __name__ == "__main__":
    cli()
